==> open-xchange-documents-ui_7.8.4.orig.tar.bz2: open-xchange-documents-ui-7.8.4/ui/apps/io.ox/office/spreadsheet/model/cellmodel.js <==
/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/model/cellmodel', [
    'io.ox/office/spreadsheet/utils/errorcode',
    'io.ox/office/spreadsheet/model/formula/tokenarray'
], function (ErrorCode, TokenArray) {

    'use strict';

    // class CellModel ========================================================

    /**
     * Light-weight representation of a defined cell in a cell collection.
     *
     * @constructor
     *
     * @param {Address} address
     *  The address of the cell model.
     *
     * @property {Address} a
     *  The address of the cell represented by this instance.
     *
     * @property {Number|String|Boolean|ErrorCode|Null} v
     *  The typed value, or formula result, of the cell. The value null is used
     *  to represent a blank cell.
     *
     * @property {String|Null} f
     *  The native formula expression of the cell (without the leading equality
     *  sign); or null for simple value cells without a formula.
     *
     * @property {Number|Null} si
     *  The index of a shared formula this cell refers to; or null if this cell
     *  is not part of a shared formula.
     *
     * @property {Range|Null} sr
     *  The bounding range of the shared formula this cell refers to, if this
     *  cell is the anchor cell of the shared formula; or null otherwise.
     *
     * @property {Range|Null} mr
     *  The bounding range of the matrix formula starting at this cell; or null
     *  otherwise.
     *
     * @property {String} s
     *  The identifier of the auto-style for the cell.
     *
     * @property {ParsedFormat} pf
     *  The parsed number format for the cell, according to the number format
     *  attribute contained in the auto-style referred by this cell model.
     *
     * @property {String|Null} d
     *  The formatted display string of the cell, according to its current
     *  value and number format. May be null, if the value cannot be formatted
     *  with the current number format. In this case, the cell renderer will
     *  show a 'railroad track' error (repeated hash marks) in the cell.
     *
     * @property {TokenArray|Null} t
     *  The token array containing the parsed formula expression stored in the
     *  property 'f' of this cell model.
     *
     * @property {String|Null} url
     *  The dynamic URL returned by a cell formula (HYPERLINK function).
     *
     * @property {CellModelVector|Null} cv
     *  Refernce to the cell model vector in the column matrix that contains
     *  this cell model; or null, if this cell model is currently not part of
     *  the column matrix.
     *
     * @property {CellModelVector|Null} rv
     *  Refernce to the cell model vector in the row matrix that contains this
     *  cell model; or null, if this cell model is currently not part of the
     *  row matrix.
     */
    function CellModel(address) {

        // 'real' model properties
        this.a = address.clone();
        this.v = null;
        this.f = null;
        this.si = null;
        this.sr = null;
        this.mr = null;
        this.s = '';

        // calculated properties, cached for performance
        this.pf = null; // MUST be initialized by caller after construction
        this.d = '';
        this.t = null;
        this.url = null;

        // references to cell vectors in the column/row matrixes
        this.cv = null;
        this.rv = null;

    } // class CellModel

    // static methods ---------------------------------------------------------

    /**
     * Compares the addresses of the passed cell models for a strict ordering,
     * e.g. for sorting cell models.
     *
     * @param {CellModel} cellModel1
     *  The first cell model whose address will be compared with the address of
     *  the second cell model.
     *
     * @param {CellModel} cellModel2
     *  The second cell model whose address will be compared with the address
     *  of the first cell model.
     *
     * @returns {Number}
     *  An integer specifying how the addresses of the passed cell models are
     *  related to each other. See method Address.compare() for details.
     */
    CellModel.compare = function (cellModel1, cellModel2) {
        return cellModel1.a.compareTo(cellModel2.a);
    };

    /**
     * Returns the column index of the passed cell model. Used in sorted model
     * arrays as sorter callback.
     *
     * @param {CellModel} cellModel
     *  The cell model to return the column index for.
     *
     * @returns {Number}
     *  The column index of the passed cell model.
     */
    CellModel.col = function (cellModel) {
        return cellModel.a[0];
    };

    /**
     * Returns the row index of the passed cell model. Used in sorted model
     * arrays as sorter callback.
     *
     * @param {CellModel} cellModel
     *  The cell model to return the row index for.
     *
     * @returns {Number}
     *  The row index of the passed cell model.
     */
    CellModel.row = function (cellModel) {
        return cellModel.a[1];
    };

    // protected methods ------------------------------------------------------

    /**
     * Creates, updates, or removes the token array stored at the property 't'
     * of this cell model, according to the current formula expression stored
     * at the property 'f'.
     *
     * @param {SheetModel} sheetModel
     *  The model of the sheet containing this cell, needed to create a new
     *  token array instance.
     */
    CellModel.prototype._updateTokenArray = function (sheetModel) {
        if (this.f) {
            if (!this.t) { this.t = new TokenArray(sheetModel, 'cell'); }
            this.t.parseFormula('op', this.f, { refAddress: this.a });
        } else {
            this.t = null;
        }
    };

    // public methods ---------------------------------------------------------

    /**
     * Returns an exact clone of this cell model.
     *
     * @param {SheetModel} sheetModel
     *  The target sheet model that will contain this cell model.
     *
     * @returns {CellModel}
     *  An exact clone of this cell model for the specified sheet model.
     */
    CellModel.prototype.clone = function (sheetModel) {
        var cellModel = new CellModel(this.a);
        cellModel.v = this.v;
        cellModel.f = this.f;
        cellModel.si = this.si;
        cellModel.sr = this.sr ? this.sr.clone() : null;
        cellModel.mr = this.mr ? this.mr.clone() : null;
        cellModel.s = this.s;
        cellModel.pf = this.pf;
        cellModel.d = this.d;
        cellModel.url = this.url;
        cellModel._updateTokenArray(sheetModel);
        return cellModel;
    };

    /**
     * Returns a unique key for the cell address of this cell model, intended
     * to be used as map key.
     *
     * @returns {String}
     *  A unique key for the cell address of this cell model.
     */
    CellModel.prototype.key = function () {
        return this.a.key();
    };

    /**
     * Returns whether this cell model is blank (no value, and no formula).
     *
     * @returns {Boolean}
     *  Whether this cell model is blank (no value, and no formula).
     */
    CellModel.prototype.isBlank = function () {
        return (this.v === null) && (this.f === null);
    };

    /**
     * Returns whether this cell model contains a number value.
     *
     * @returns {Boolean}
     *  Whether this cell model contains a number value.
     */
    CellModel.prototype.isNumber = function () {
        return typeof this.v === 'number';
    };

    /**
     * Returns whether this cell model contains a string value.
     *
     * @returns {Boolean}
     *  Whether this cell model contains a string value.
     */
    CellModel.prototype.isText = function () {
        return typeof this.v === 'string';
    };

    /**
     * Returns whether this cell model contains a boolean value.
     *
     * @returns {Boolean}
     *  Whether this cell model contains a boolean value.
     */
    CellModel.prototype.isBoolean = function () {
        return typeof this.v === 'boolean';
    };

    /**
     * Returns whether the cell contains an error code.
     *
     * @returns {Boolean}
     *  Whether the cell contains an error code.
     */
    CellModel.prototype.isError = function () {
        return this.v instanceof ErrorCode;
    };

    /**
     * Returns whether this cell model represents a normal formula cell, i.e.
     * the cell contains a formula expression, but is neither part of a shared
     * formula nor a matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model represents a normal formula cell.
     */
    CellModel.prototype.isNormalFormula = function () {
        return !!this.f && (this.si === null) && !this.mr;
    };

    /**
     * Returns whether this cell model is part of a shared formula, regardless
     * whether the cell is the anchor of the shared formula.
     *
     * @returns {Boolean}
     *  Whether this cell model is part of a shared formula.
     */
    CellModel.prototype.isSharedFormula = function () {
        return this.si !== null;
    };

    /**
     * Returns whether this cell model contains a formula, either as normal
     * formula cell, or as part of a shared formula, or as the anchor cell of a
     * matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model represents a formula cell.
     */
    CellModel.prototype.isAnyFormula = function () {
        return !!this.f || (this.si !== null) || !!this.mr;
    };

    /**
     * Returns whether this cell model is the anchor cell of a shared formula.
     *
     * @returns {Boolean}
     *  Whether this cell model is the anchor cell of a shared formula.
     */
    CellModel.prototype.isSharedAnchor = function () {
        return !!this.sr;
    };

    /**
     * Returns whether this cell model is the anchor cell of a matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model is the anchor cell of a matrix formula.
     */
    CellModel.prototype.isMatrixAnchor = function () {
        return !!this.mr;
    };

    /**
     * Returns whether this cell model contains an actual formula expression,
     * either as normal formula cell, or as anchor cell of a shared formula or
     * matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model contains an actual formula expression.
     */
    CellModel.prototype.isAnyAnchor = function () {
        return !!this.f || !!this.sr || !!this.mr;
    };

    /**
     * Returns the formula expression of this cell model.
     *
     * @param {String} grammarId
     *  The identifier of the formula grammar for the formula expression. See
     *  SpreadsheetDocument.getFormulaGrammar() for more details.
     *
     * @param {Address} [targetAddress]
     *  The target address used to relocate the formula expression. If omitted,
     *  the formula will be generated for the own address of this cell model.
     *
     * @returns {String|Null}
     *  The formula expression of this cell model; or null, if the cell model
     *  does not contain a formula expression.
     */
    CellModel.prototype.getFormula = function (grammarId, targetAddress) {
        return !this.t ? null :
            // shortcut for operation grammar without relocation: return the 'f' property directly
            ((grammarId === 'op') && !targetAddress) ? this.f :
            this.t.getFormula(grammarId, { refAddress: this.a, targetAddress: targetAddress || this.a });
    };

    /**
     * Changes the value of this cell model, and returns whether the cell value
     * has changed.
     *
     * @param {Number|String|Boolean|ErrorCode|Null} value
     *  The new value for the cell.
     *
     * @returns {Boolean}
     *  Whether the cell value has changed.
     */
    CellModel.prototype.setValue = function (value) {
        if (this.v === value) { return false; }
        this.v = value;
        return true;
    };

    /**
     * Changes the formula expression of this cell model, and returns whether
     * the formula expression has changed.
     *
     * @param {SheetModel} sheetModel
     *  The sheet model containing this cell model, needed to create a new
     *  token array instance on demand.
     *
     * @param {String|Null} formula
     *  The new formula expression for the cell. The value null will delete the
     *  formula expression, and the cell becomes a simple value cell.
     *
     * @param {Boolean} [importing=false]
     *  If set to true, the token array of the cell model will NOT be created
     *  (this will be done later during post-processing phase of the import
     *  process to improve performance).
     *
     * @returns {Boolean}
     *  Whether the formula expression has changed.
     */
    CellModel.prototype.setFormula = function (sheetModel, formula, importing) {
        formula = formula || null;
        if (this.f === formula) { return false; }
        this.f = formula;
        if (!importing) { this._updateTokenArray(sheetModel); }
        return true;
    };

    /**
     * Changes the index of the shared formula this cell refers to.
     *
     * @param {Number|Null} si
     *  The index of a shared formula, or null to remove the existing index of
     *  a shared formula from this cell model.
     *
     * @returns {Boolean}
     *  Whether the index of the shared formula has changed.
     */
    CellModel.prototype.setSharedIndex = function (si) {
        if (this.si === si) { return false; }
        this.si = si;
        return true;
    };

    /**
     * Changes the bounding range of the shared formula this cell refers to.
     *
     * @param {Range|Null} range
     *  The new bounding range of a shared formula, or null to remove the
     *  bounding range from this cell model.
     *
     * @returns {Boolean}
     *  Whether the bounding range of the shared formula has changed.
     */
    CellModel.prototype.setSharedRange = function (range) {
        if ((!this.sr && !range) || (this.sr && range && this.sr.equals(range))) { return false; }
        this.sr = range ? range.clone() : null;
        return true;
    };

    /**
     * Changes the bounding range of the matrix formula starting at this cell.
     *
     * @param {Range|Null} range
     *  The new bounding range of a matrix formula, or null to remove the
     *  bounding range from this cell model.
     *
     * @returns {Boolean}
     *  Whether the bounding range of the matrix formula has changed.
     */
    CellModel.prototype.setMatrixRange = function (range) {
        if ((!this.mr && !range) || (this.mr && range && this.mr.equals(range))) { return false; }
        this.mr = range ? range.clone() : null;
        return true;
    };

    /**
     * Changes the dynamic URL of this cell model, and returns whether the cell
     * URL has changed.
     *
     * @param {String|Null} url
     *  The new dynamic URL for the cell, or null to remove the current URL.
     *
     * @returns {Boolean}
     *  Whether the dynamic cell URL has changed.
     */
    CellModel.prototype.setURL = function (url) {
        if (this.url === url) { return false; }
        this.url = url;
        return true;
    };

    // exports ================================================================

    return CellModel;

});

==> open-xchange-documents-ui_7.10.1.orig.tar.bz2: open-xchange-documents-ui-7.10.1/ui/apps/io.ox/office/spreadsheet/model/cellmodel.js <==
/**
 * This work is provided under the terms of the CREATIVE COMMONS PUBLIC
 * LICENSE. This work is protected by copyright and/or other applicable
 * law. Any use of the work other than as authorized under this license
 * or copyright law is prohibited.
 *
 * http://creativecommons.org/licenses/by-nc-sa/2.5/
 *
 * Copyright (C) 2016 OX Software GmbH
 * Mail: info@open-xchange.com
 *
 * @author Daniel Rentz <daniel.rentz@open-xchange.com>
 */

define('io.ox/office/spreadsheet/model/cellmodel', [
    'io.ox/office/spreadsheet/utils/sheetutils'
], function (SheetUtils) {

    'use strict';

    // convenience shortcuts
    var ErrorCode = SheetUtils.ErrorCode;
    var Address = SheetUtils.Address;

    // class CellModel ========================================================

    /**
     * Light-weight representation of a defined cell in a cell collection.
     *
     * @constructor
     *
     * @param {Address} address
     *  The address of the cell model.
     *
     * @property {Address} a
     *  The address of the cell represented by this instance.
     *
     * @property {Number|String|Boolean|ErrorCode|Null} v
     *  The typed value, or formula result, of the cell. The value null is used
     *  to represent a blank cell.
     *
     * @property {String|Null} f
     *  The native formula expression of the cell (without the leading equality
     *  sign); or null for simple value cells without a formula.
     *
     * @property {Number|Null} si
     *  The index of a shared formula this cell refers to; or null if this cell
     *  is not part of a shared formula.
     *
     * @property {Range|Null} sr
     *  The bounding range of the shared formula this cell refers to, if this
     *  cell is the anchor cell of the shared formula; or null otherwise.
     *
     * @property {Range|Null} mr
     *  The bounding range of the matrix formula starting at this cell; or null
     *  otherwise.
     *
     * @property {String} s
     *  The identifier of the auto-style for the cell.
     *
     * @property {ParsedFormat} pf
     *  The parsed number format for the cell, according to the number format
     *  attribute contained in the auto-style referred by this cell model.
     *
     * @property {String|Null} d
     *  The formatted display string of the cell, according to its current
     *  value and number format. May be null, if the value cannot be formatted
     *  with the current number format. In this case, the cell renderer will
     *  show a 'railroad track' error (repeated hash marks) in the cell.
     *
     * @property {TokenArray|Null} t
     *  The token array containing the parsed formula expression stored in the
     *  property 'f' of this cell model.
     *
     * @property {String|Null} url
     *  The dynamic URL returned by a cell formula (HYPERLINK function).
     *
     * @property {CellModelVector|Null} cv
     *  Refernce to the cell model vector in the column matrix that contains
     *  this cell model; or null, if this cell model is currently not part of
     *  the column matrix.
     *
     * @property {CellModelVector|Null} rv
     *  Refernce to the cell model vector in the row matrix that contains this
     *  cell model; or null, if this cell model is currently not part of the
     *  row matrix.
     */
    function CellModel(address) {

        // 'real' model properties
        this.a = address.clone();
        this.v = null;
        this.f = null;
        this.si = null;
        this.sr = null;
        this.mr = null;
        this.s = '';

        // calculated properties, cached for performance
        this.pf = null; // MUST be initialized by caller after construction
        this.d = '';
        this.t = null;
        this.url = null;

        // references to cell vectors in the column/row matrixes
        this.cv = null;
        this.rv = null;

    } // class CellModel

    // static methods ---------------------------------------------------------

    /**
     * Compares the addresses of the passed cell models for a strict horizontal
     * ordering, e.g. for sorting cell models.
     *
     * @param {CellModel} cellModel1
     *  The first cell model whose address will be compared with the address of
     *  the second cell model.
     *
     * @param {CellModel} cellModel2
     *  The second cell model whose address will be compared with the address
     *  of the first cell model.
     *
     * @returns {Number}
     *  An integer specifying how the addresses of the passed cell models are
     *  related to each other. See method Address.compare() for details.
     */
    CellModel.compare = function (cellModel1, cellModel2) {
        return Address.compare(cellModel1.a, cellModel2.a);
    };

    /**
     * Compares the addresses of the passed cell models for a strict vertical
     * ordering, e.g. for sorting cell models.
     *
     * @param {CellModel} cellModel1
     *  The first cell model whose address will be compared with the address of
     *  the second cell model.
     *
     * @param {CellModel} cellModel2
     *  The second cell model whose address will be compared with the address
     *  of the first cell model.
     *
     * @returns {Number}
     *  An integer specifying how the addresses of the passed cell models are
     *  related to each other. See method Address.compareVert() for details.
     */
    CellModel.compareVert = function (cellModel1, cellModel2) {
        return Address.compareVert(cellModel1.a, cellModel2.a);
    };

    /**
     * Compares the column indexes of the passed cell models.
     *
     * @param {CellModel} cellModel1
     *  The first cell model whose address will be compared with the address of
     *  the second cell model.
     *
     * @param {CellModel} cellModel2
     *  The second cell model whose address will be compared with the address
     *  of the first cell model.
     *
     * @returns {Number}
     *  An integer specifying how the addresses of the passed cell models are
     *  related to each other. See method Address.compareCol() for details.
     */
    CellModel.compareCol = function (cellModel1, cellModel2) {
        return Address.compareCol(cellModel1.a, cellModel2.a);
    };

    /**
     * Compares the row indexes of the passed cell models.
     *
     * @param {CellModel} cellModel1
     *  The first cell model whose address will be compared with the address of
     *  the second cell model.
     *
     * @param {CellModel} cellModel2
     *  The second cell model whose address will be compared with the address
     *  of the first cell model.
     *
     * @returns {Number}
     *  An integer specifying how the addresses of the passed cell models are
     *  related to each other. See method Address.compareRow() for details.
     */
    CellModel.compareRow = function (cellModel1, cellModel2) {
        return Address.compareRow(cellModel1.a, cellModel2.a);
    };

    /**
     * Returns the column index of the passed cell model. Used in sorted model
     * arrays as sorter callback.
     *
     * @param {CellModel} cellModel
     *  The cell model to return the column index for.
     *
     * @returns {Number}
     *  The column index of the passed cell model.
     */
    CellModel.col = function (cellModel) {
        return cellModel.a[0];
    };

    /**
     * Returns the row index of the passed cell model. Used in sorted model
     * arrays as sorter callback.
     *
     * @param {CellModel} cellModel
     *  The cell model to return the row index for.
     *
     * @returns {Number}
     *  The row index of the passed cell model.
     */
    CellModel.row = function (cellModel) {
        return cellModel.a[1];
    };

    // protected methods ------------------------------------------------------

    /**
     * Creates, updates, or removes the token array stored at the property 't'
     * of this cell model, according to the current formula expression stored
     * at the property 'f'.
     *
     * @param {SheetModel} sheetModel
     *  The model of the sheet containing this cell, needed to create a new
     *  token array instance.
     */
    CellModel.prototype._updateTokenArray = function (sheetModel) {
        if (this.f) {
            if (!this.t) { this.t = sheetModel.createCellTokenArray(); }
            this.t.parseFormula('op', this.f, { refAddress: this.a });
        } else {
            this.t = null;
        }
    };

    // public methods ---------------------------------------------------------

    /**
     * Returns an exact clone of this cell model.
     *
     * @param {SheetModel} sheetModel
     *  The target sheet model that will contain this cell model.
     *
     * @returns {CellModel}
     *  An exact clone of this cell model for the specified sheet model.
     */
    CellModel.prototype.clone = function (sheetModel) {
        var cellModel = new CellModel(this.a);
        cellModel.v = this.v;
        cellModel.f = this.f;
        cellModel.si = this.si;
        cellModel.sr = this.sr ? this.sr.clone() : null;
        cellModel.mr = this.mr ? this.mr.clone() : null;
        cellModel.s = this.s;
        cellModel.pf = this.pf;
        cellModel.d = this.d;
        cellModel.url = this.url;
        cellModel._updateTokenArray(sheetModel);
        return cellModel;
    };

    /**
     * Returns a unique key for the cell address of this cell model, intended
     * to be used as map key.
     *
     * @returns {String}
     *  A unique key for the cell address of this cell model.
     */
    CellModel.prototype.key = function () {
        return this.a.key();
    };

    /**
     * Returns whether this cell model is blank (no value, and no formula).
     *
     * @returns {Boolean}
     *  Whether this cell model is blank (no value, and no formula).
     */
    CellModel.prototype.isBlank = function () {
        return (this.v === null) && (this.f === null);
    };

    /**
     * Returns whether this cell model contains a number value.
     *
     * @returns {Boolean}
     *  Whether this cell model contains a number value.
     */
    CellModel.prototype.isNumber = function () {
        return typeof this.v === 'number';
    };

    /**
     * Returns whether this cell model contains a string value.
     *
     * @returns {Boolean}
     *  Whether this cell model contains a string value.
     */
    CellModel.prototype.isText = function () {
        return typeof this.v === 'string';
    };

    /**
     * Returns whether this cell model contains a boolean value.
     *
     * @returns {Boolean}
     *  Whether this cell model contains a boolean value.
     */
    CellModel.prototype.isBoolean = function () {
        return typeof this.v === 'boolean';
    };

    /**
     * Returns whether the cell contains an error code.
     *
     * @returns {Boolean}
     *  Whether the cell contains an error code.
     */
    CellModel.prototype.isError = function () {
        return this.v instanceof ErrorCode;
    };

    /**
     * Returns whether this cell model represents a normal formula cell, i.e.
     * the cell contains a formula expression, but is neither part of a shared
     * formula nor a matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model represents a normal formula cell.
     */
    CellModel.prototype.isNormalFormula = function () {
        return !!this.f && (this.si === null) && !this.mr;
    };

    /**
     * Returns whether this cell model is part of a shared formula, regardless
     * whether the cell is the anchor of the shared formula.
     *
     * @returns {Boolean}
     *  Whether this cell model is part of a shared formula.
     */
    CellModel.prototype.isSharedFormula = function () {
        return this.si !== null;
    };

    /**
     * Returns whether this cell model contains a formula, either as normal
     * formula cell, or as part of a shared formula, or as the anchor cell of a
     * matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model represents a formula cell.
     */
    CellModel.prototype.isAnyFormula = function () {
        return !!this.f || (this.si !== null) || !!this.mr;
    };

    /**
     * Returns whether this cell model is the anchor cell of a shared formula.
     *
     * @returns {Boolean}
     *  Whether this cell model is the anchor cell of a shared formula.
     */
    CellModel.prototype.isSharedAnchor = function () {
        return !!this.sr;
    };

    /**
     * Returns whether this cell model is the anchor cell of a matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model is the anchor cell of a matrix formula.
     */
    CellModel.prototype.isMatrixAnchor = function () {
        return !!this.mr;
    };

    /**
     * Returns whether this cell model contains an actual formula expression,
     * either as normal formula cell, or as anchor cell of a shared formula or
     * matrix formula.
     *
     * @returns {Boolean}
     *  Whether this cell model contains an actual formula expression.
     */
    CellModel.prototype.isAnyAnchor = function () {
        return !!this.f || !!this.sr || !!this.mr;
    };

    /**
     * Returns the formula expression of this cell model.
     *
     * @param {String} grammarId
     *  The identifier of the formula grammar for the formula expression. See
     *  SpreadsheetDocument.getFormulaGrammar() for more details.
     *
     * @param {Address} [targetAddress]
     *  The target address used to relocate the formula expression. If omitted,
     *  the formula will be generated for the own address of this cell model.
     *
     * @returns {String|Null}
     *  The formula expression of this cell model; or null, if the cell model
     *  does not contain a formula expression.
     */
    CellModel.prototype.getFormula = function (grammarId, targetAddress) {
        return !this.t ? null :
            // shortcut for operation grammar without relocation: return the 'f' property directly
            ((grammarId === 'op') && !targetAddress) ? this.f :
            this.t.getFormula(grammarId, { refAddress: this.a, targetAddress: targetAddress || this.a });
    };

    /**
     * Changes the value of this cell model, and returns whether the cell value
     * has changed.
     *
     * @param {Number|String|Boolean|ErrorCode|Null} value
     *  The new value for the cell.
     *
     * @returns {Boolean}
     *  Whether the cell value has changed.
     */
    CellModel.prototype.setValue = function (value) {
        if (this.v === value) { return false; }
        this.v = value;
        return true;
    };

    /**
     * Changes the formula expression of this cell model, and returns whether
     * the formula expression has changed.
     *
     * @param {SheetModel} sheetModel
     *  The sheet model containing this cell model, needed to create a new
     *  token array instance on demand.
     *
     * @param {String|Null} formula
     *  The new formula expression for the cell. The value null will delete the
     *  formula expression, and the cell becomes a simple value cell.
     *
     * @param {Boolean} [importing=false]
     *  If set to true, the token array of the cell model will NOT be created
     *  (this will be done later during post-processing phase of the import
     *  process to improve performance).
     *
     * @returns {Boolean}
     *  Whether the formula expression has changed.
     */
    CellModel.prototype.setFormula = function (sheetModel, formula, importing) {
        formula = formula || null;
        if (this.f === formula) { return false; }
        this.f = formula;
        if (!importing) { this._updateTokenArray(sheetModel); }
        return true;
    };

    /**
     * Changes the index of the shared formula this cell refers to.
     *
     * @param {Number|Null} si
     *  The index of a shared formula, or null to remove the existing index of
     *  a shared formula from this cell model.
     *
     * @returns {Boolean}
     *  Whether the index of the shared formula has changed.
     */
    CellModel.prototype.setSharedIndex = function (si) {
        if (this.si === si) { return false; }
        this.si = si;
        return true;
    };

    /**
     * Changes the bounding range of the shared formula this cell refers to.
     *
     * @param {Range|Null} range
     *  The new bounding range of a shared formula, or null to remove the
     *  bounding range from this cell model.
     *
     * @returns {Boolean}
     *  Whether the bounding range of the shared formula has changed.
     */
    CellModel.prototype.setSharedRange = function (range) {
        if ((!this.sr && !range) || (this.sr && range && this.sr.equals(range))) { return false; }
        this.sr = range ? range.clone() : null;
        return true;
    };

    /**
     * Changes the bounding range of the matrix formula starting at this cell.
     *
     * @param {Range|Null} range
     *  The new bounding range of a matrix formula, or null to remove the
     *  bounding range from this cell model.
     *
     * @returns {Boolean}
     *  Whether the bounding range of the matrix formula has changed.
     */
    CellModel.prototype.setMatrixRange = function (range) {
        if ((!this.mr && !range) || (this.mr && range && this.mr.equals(range))) { return false; }
        this.mr = range ? range.clone() : null;
        return true;
    };

    /**
     * Changes the dynamic URL of this cell model, and returns whether the cell
     * URL has changed.
     *
     * @param {String|Null} url
     *  The new dynamic URL for the cell, or null to remove the current URL.
     *
     * @returns {Boolean}
     *  Whether the dynamic cell URL has changed.
     */
    CellModel.prototype.setURL = function (url) {
        if (this.url === url) { return false; }
        this.url = url;
        return true;
    };

    // exports ================================================================

    return CellModel;

});
